# -*- coding: utf-8 -*-

from .common import load_file
from importlib import import_module


class Record:

    @staticmethod
    def create_from_file(file, password_generate=False):
        record_data = load_file(file)

        if record_data.get("kind") != "KeeperRecord":
            raise ValueError(".kind is not 'KeeperRecord'")

        version = record_data.get("version")
        if version is None or version == "":
            raise ValueError(".version is missing or blank")

        mod = import_module(f"keeper_secrets_manager_helper.{version}.record")
        return getattr(mod, "Record").create_from_data(record_data, password_generate=password_generate)

    def __init__(self, version):

        self.version = version

        try:
            self.record_mod = import_module(f"keeper_secrets_manager_helper.{self.version}.record")
            self.parser_mod = import_module(f"keeper_secrets_manager_helper.{self.version}.parser")
            self.record_type_mod = import_module(f"keeper_secrets_manager_helper.{self.version}.record_type")
        except ImportError as err:
            raise Exception(f"Version {self.version} is not supported: " + str(err))

    def create_from_field_list(self, record_type, fields, title=None, notes=None,
                               password_generate=None, password_complexity=None):
        return [getattr(self.record_mod, "Record")(
            record_type=record_type,
            title=title,
            notes=notes,
            fields=fields,
            password_generate=password_generate,
            password_complexity=password_complexity
        )]

    def create_from_field_args(self, **kwargs):
        kwargs["fields"] = getattr(self.parser_mod, "Parser")().parse_field(kwargs.get("field_args"))
        return [getattr(self.record_mod, "Record")(**kwargs)]

    def get_template(self, record_type, output_format, title=None, notes=None):
        get_class_by_type = getattr(self.record_type_mod, "get_class_by_type")
        record_type = get_class_by_type(record_type)()
        return record_type.generate_template(output_format=output_format, title=title, notes=notes)

    def get_template_list(self):
        return getattr(self.record_type_mod, "get_record_type_list")()



# -*- coding: utf-8 -*-

from enum import Enum
import json


class FieldSectionEnum(Enum):
    STANDARD = "f"
    CUSTOM = "c"


class Field:

    complete_key = "_complete"

    def __init__(self, **kwargs):

        self.initial_value_was_json = None

        # If an array of values is passed in, then the field is complete.
        self.is_complete = False

        self.type = kwargs.pop("type", None)
        self.field_section = kwargs.pop("field_section", None)
        self.label = kwargs.pop("label", None)
        self.allow_multiple = kwargs.pop("allow_multiple", None)
        self._value = None
        self.value = kwargs.pop("value", None)

        value_key = kwargs.pop("value_key", None)
        if value_key is not None and isinstance(self.value, dict) is False:
            self.initial_value_was_json = False
            self.value = {
                value_key: self.value
            }
        self.value_key = value_key

        # Default at the way way way end.
        self.index = kwargs.pop("index", 1_000_000)

        self.group_key = kwargs.pop("group_key", None)

        self.extra = kwargs

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        if val is not None:
            # If the value JSON? Decode it if it is.
            if isinstance(val, str) is True:
                try:
                    val = json.loads(val)

                    # If the initial value was JSON, then flag it was and mark field as complete.
                    if self.initial_value_was_json is None:
                        if isinstance(val, list) is True:
                            self.is_complete = True
                            # Is complete, no need to flag any dictionary as comelete
                        elif isinstance(val, dict) is True:
                            val[Field.complete_key] = True
                        self.initial_value_was_json = True
                except json.JSONDecodeError:
                    self.initial_value_was_json = False
            if val == "":
                val = None
        self._value = val

    def __str__(self):
        return f'Field(type={self.type}, field_section={self.field_section}, label={self.label}, '\
               f'value_key={self.value_key}, value={self.value}, index={self.index}, extra={self.extra}, '\
               f'group_key={self.group_key}, initial_value_was_json={self.initial_value_was_json}'

    def add_extra(self, key, value):
        if hasattr(self, key) is False:
            self.extra[key] = value

    @staticmethod
    def field_key(field_type, label):
        return "{}/{}".format(field_type, label)

    def instance_field_key(self, label=None):
        return self.field_key(self.type, label if self.label is None else self.label)

    def can_add_key_value(self):
        return isinstance(self.value, dict) is True or self.group_key is not None

    def to_dict(self):
        data = {
            "type": self.type,
            "label": self.label,
            "value": self.value,
        }
        return {**data, **self.extra}



# -*- coding: utf-8 -*-

from keeper_secrets_manager_helper.field import Field, FieldSectionEnum
from keeper_secrets_manager_helper.common import load_file
from keeper_secrets_manager_helper.v3.record_type import get_class_by_type as get_record_type_class
from keeper_secrets_manager_helper.v3.field_type import get_class_by_type as get_field_type_class
from importlib import import_module


class Record:

    @staticmethod
    def create_from_file(file, password_generate=False):
        record_data = load_file(file)
        return Record.create_from_data(record_data, password_generate=password_generate)

    @staticmethod
    def create_from_data(record_data, password_generate=False):

        records = []

        if record_data.get("version") != "v3":
            raise ValueError(".version is not 'v3'")
        if record_data.get("kind") != "KeeperRecord":
            raise ValueError(".kind is not 'KeeperRecord'")
        data = record_data.get("data")
        if data is None:
            raise ValueError(".data[] is missing")
        if isinstance(data, list) is False:
            raise ValueError(".data[] is not an array")

        record_count = 0
        for record_item in data:
            record_type = record_item.get("recordType", record_item.get("record_type"))
            if record_type is None or record_type == "":
                raise ValueError(f".data[{record_count}].recordType is missing or blank")
            title = record_item.get("title")
            if title is None or title == "":
                raise ValueError(f".data[{record_count}].title is missing or blank")

            record = Record(
                record_type=record_type,
                title=title,
                notes=record_item.get("notes"),
                password_generate=password_generate
            )

            all_fields = []

            fields = record_item.get("fields")
            if fields is None:
                raise ValueError(f".data[{record_count}].fields[] is missing")
            if isinstance(fields, list) is False:
                raise ValueError(f".data[{record_count}].fields[] is not an array")

            for field_item in fields:
                field = Field(
                    type=field_item.get("type"),
                    field_section=FieldSectionEnum.STANDARD,
                    label=field_item.get("label"),
                    value=field_item.get("value"),
                )
                all_fields.append(field)

            custom_fields = record_item.get("customFields", record_item.get("custom_fields"))
            if custom_fields is not None:
                if isinstance(custom_fields, list) is False:
                    raise ValueError(f".data[{record_count}].fields[] is not an array")

                for field_item in custom_fields:
                    field = Field(
                        type=field_item.get("type"),
                        field_section=FieldSectionEnum.CUSTOM,
                        label=field_item.get("label"),
                        value=field_item.get("value"),
                    )
                    all_fields.append(field)

            record.add_fields(all_fields)
            record.build_record()
            records.append(record)

        return records

    def __init__(self, *args, **kwargs):

        # If there is an arg, then assume it's a dictionary with record data.
        if len(args) > 0:
            pass

        self.record_type = kwargs.get("record_type")
        self.title = kwargs.get("title")
        self.notes = kwargs.get("notes")
        self.fields = []
        self.custom_fields = []

        if self.record_type is None or self.record_type == "":
            raise ValueError("record_type is missing or blank.")

        try:
            record_type = get_record_type_class(self.record_type)()

            # Make a quick lookup for the standard fields.
            self._valid_fields = [{"type": x.get("type"), "label": x.get("label"), "has_value": False}
                                  for x in record_type.get_standard_fields()]
        except ImportError as err:
            raise ValueError(err)

        if self.title is None or self.title == "":
            raise ValueError("title is missing or blank.")

        # The fields are mapped here in an attempt to make unique fields.
        self._fields = {
            FieldSectionEnum.STANDARD: {},
            FieldSectionEnum.CUSTOM: {}
        }

        self.password_generate = kwargs.get("password_generate", False)
        self.password_complexity = kwargs.get("password_complexity", None)

        self.valid_fields = []

        # All the fields (standard/custom) to be passed in with the constructor.
        fields = kwargs.get("fields")
        if fields is not None:
            self.add_fields(fields)
            self.build_record()

    def _add_new_field(self, field, field_key, group_key):
        # Count the number of keys in the dictionary and use that for an index. That will be used determine
        # the order.
        field.index = len(self._fields[field.field_section])

        # If the group key is not None, then convert the value to an array.
        if group_key is not None and isinstance(field.value, list) is False:
            field.value = [field.value]

        self._fields[field.field_section][field_key] = field

    def _is_valid_standard_field(self, field_type):
        for item in self._valid_fields:
            if item.get("type") == field_type and item.get("has_value") is False:
                return True
        return False

    def _flag_standard_field_used(self, field_type):
        for item in self._valid_fields:
            if item.get("type") == field_type and item.get("has_value") is False:
                item["has_value"] = True
                break

    def _get_label_for_standard_field(self, field_type):
        for item in self._valid_fields:
            if item.get("type") == field_type and item.get("has_value") is False:
                return item.get("label")
        return None

    def add_fields(self, fields):
        if isinstance(fields, list) is False:
            fields = [fields]

        for field in fields:

            if isinstance(field, Field) is False:
                raise ValueError("The method add_field requires instance(s) of Field")

            #
            label = None
            if field.field_section == FieldSectionEnum.STANDARD:
                label = self._get_label_for_standard_field(field.type)

            field_key = field.instance_field_key(label=label)
            group_key = field.group_key

            # Does this key already exists? And can we add values to the dictionary value?
            if field_key in self._fields[field.field_section] and field.can_add_key_value():

                # If out value is a string we should not be in here.
                if isinstance(field.value, str) is True:
                    raise ValueError(f"The {field.type} is a string. If JSON check to see if JSON is valid.")

                # Get the existing field and copy any values in it's dictionary into the existing.
                existing_field = self._fields[field.field_section][field_key]

                # If the field is completely set
                if existing_field.is_complete is True and existing_field.field_section == FieldSectionEnum.STANDARD:
                    raise ValueError("Attempting to set a standard field that has already been set.")

                # The existing field is complete and a custom field, so add
                if existing_field.is_complete is True:
                    raise ValueError("Cannot add this field due to it not being unique. To make unique add a label to "
                                     "the field or make sure the label is not being duplicated.")

                # If the existing_field is JSON and the current field is JSON, then add to existing. This allows
                # the value to be set with multiple objects.
                if existing_field.initial_value_was_json and field.initial_value_was_json:
                    if isinstance(existing_field.value, dict) is True:
                        existing_field.value = [existing_field.value]
                    if isinstance(field.value, list) is True:
                        for item in field.value:
                            existing_field.value.append(item)
                    else:
                        existing_field.value.append(field.value)
                    continue

                for k, v in field.value.items():

                    # If tke group key is set. The value can be multiple dictionaries that have a specific key
                    # which indicates its uniqueness. If that key does not exist, values can be inserted into the
                    # last dictionary in the list. If does exists, then a new dictionary is created.
                    if group_key is not None:
                        found_a_place = False
                        for item in existing_field.value:
                            if group_key not in item and item.get(Field.complete_key) is not True:
                                item[k] = v
                                found_a_place = True
                            else:
                                item[Field.complete_key] = True
                        if found_a_place is False and isinstance(existing_field.value, list) is True:
                            new_object = {k: v}
                            existing_field.value.append(new_object)
                    elif isinstance(existing_field.value, dict) is True:
                        existing_field.value[k] = v

            # Else we are creating a new entry.
            else:
                # Standard fields are defined. Don't insert a field that doesn't belong.
                if field.field_section == FieldSectionEnum.STANDARD:
                    if self._is_valid_standard_field(field.type):
                        self._flag_standard_field_used(field.type)
                    else:
                        raise ValueError(f"The standard fields do not have a '{field.type}' "
                                         "field type or they all have values.")

                self._add_new_field(field, field_key, group_key)

    @staticmethod
    def _copy_record_type_settings(field_obj, standard_field):
        # Copy extra values from the record type schema to the field. These are unique field type params like
        # required, enforce_generation and complexity.
        for key, value in standard_field.items():
            field_obj.add_extra(key, value)

    def _get_standard_fields(self, record_type):

        # Add the standard fields in the order defined by record type schema.

        fields_list = []
        # Get a list of standard fields in the Record Type.
        for standard_field in record_type.get_standard_fields():
            # First check if we have a key with a label, if it exists, and then use that.
            field_key = Field.field_key(standard_field.get("type"), standard_field.get("label"))
            if field_key in self._fields[FieldSectionEnum.STANDARD]:
                field_obj = self._fields[FieldSectionEnum.STANDARD][field_key]
                self._copy_record_type_settings(field_obj, standard_field)
                fields_list.append(field_obj)
            else:
                # Find the field by it's field type.
                field_key = Field.field_key(standard_field.get("type"), None)
                if field_key in self._fields[FieldSectionEnum.STANDARD]:
                    field_obj = self._fields[FieldSectionEnum.STANDARD][field_key]
                    self._copy_record_type_settings(field_obj, standard_field)
                    fields_list.append(field_obj)
                else:
                    # If nothing exists, make an empty field for the field type
                    field_obj = Field(
                        type=standard_field.get("type"),
                        field_section=FieldSectionEnum.STANDARD,
                        value=None
                    )
                    self._copy_record_type_settings(field_obj, standard_field)
                    fields_list.append(field_obj)

        return fields_list

    def _get_custom_fields(self):

        def get_index_key(obj):
            return obj.index

        # Add the custom fields in the order they were added.
        fields_list = [self._fields[FieldSectionEnum.CUSTOM][x] for x in self._fields[FieldSectionEnum.CUSTOM]]
        fields_list.sort(key=get_index_key)
        return fields_list

    @staticmethod
    def _remove_private_keys(obj):
        """
        The value might contain dictionaries what contain private key. This will remove any that exists. Right
        now it's just one.
        """
        if isinstance(obj, list):
            for item in obj:
                Record._remove_private_keys(item)
        elif isinstance(obj, dict):
            obj.pop(Field.complete_key, None)

    def build_record(self):

        record_type = get_record_type_class(self.record_type)()

        # Take all the standard fields from the user's input and populate the field type to validate it. Then
        # the dictionary used in the V3 records for a field to the list.
        self.fields = []
        for field in self._get_standard_fields(record_type):
            field_type_kwargs = field.to_dict()
            self._remove_private_keys(field_type_kwargs.get("value"))
            field_type_kwargs["password_generate"] = self.password_generate
            if self.password_complexity is not None:
                field_type_kwargs["complexity"] = self.password_complexity
            field_type_obj = get_field_type_class(field.type)(**field_type_kwargs)
            self.fields.append(field_type_obj.to_dict())

        # Do the same with the custom fields.
        self.custom_fields = []
        for field in self._get_custom_fields():
            field_type_kwargs = field.to_dict()
            self._remove_private_keys(field_type_kwargs.get("value"))
            field_type_kwargs["password_generate"] = self.password_generate
            if self.password_complexity is not None:
                field_type_kwargs["complexity"] = self.password_complexity
            field_type_obj = get_field_type_class(field.type)(**field_type_kwargs)
            self.custom_fields.append(field_type_obj.to_dict())

    def get_record_create_obj(self):
        try:
            # Make sure the classes we need are in the KSM Python SDK.
            mod = import_module("keeper_secrets_manager_core.dto.dtos")
            if hasattr(mod, "RecordCreate") is False:
                raise ImportError("Cannot find the RecordCreate in the KSM Python SDK. Please update the SDK.")
            record_field_class = getattr(mod, "RecordField")
            if record_field_class is None:
                raise ImportError("Cannot find the RecordField in the KSM Python SDK. Please update the SDK.")

            # Make an instance of the SDK's RecordCreate
            new_record = getattr(mod, "RecordCreate")(
                record_type=self.record_type,
                title=self.title
            )

            # Add the standard fields thru RecordField constructor
            record_field = []
            for field in self.fields:

                # Translate dictionary to RecordField
                field["field_type"] = field.pop("type")

                # V3 does take complexity or enforceGeneration
                field.pop("complexity", None)
                field.pop("enforceGeneration", None)

                record_field.append(record_field_class(**field))
            new_record.fields = record_field

            # Add the custom fields thru RecordField constructor
            record_field = []
            for field in self.custom_fields:
                # Translate dictionary to RecordField
                field["field_type"] = field.pop("type")

                # V3 does take complexity or enforceGeneration
                field.pop("complexity", None)
                field.pop("enforceGeneration", None)

                record_field.append(record_field_class(**field))
            new_record.custom = record_field

            # Add the notes
            new_record.notes = self.notes

        except ImportError as _:
            raise Exception("Cannot build a CreateRecord instance. Cannot find the KSM Python SDK.")

        return new_record



# -*- coding: utf-8 -*-

import re
import json
import inspect
import sys
import random
from importlib import import_module
import keeper_secrets_manager_helper.format
from keeper_secrets_manager_helper.v3.enum import BaseEnum, PhoneTypeEnum, CountryEnum, AccountTypeEnum


UID_REGEX = r'^[a-zA-Z0-9\-_]{22}$'

field_map = {}


# Find all the field type classes and create a map from the the camelcase name to the FieldType class
def get_field_type_map():
    global field_map
    if len(field_map) == 0:
        for item in inspect.getmembers(sys.modules[__name__], inspect.isclass):
            mod_class = getattr(sys.modules[__name__], item[0])
            if mod_class is not None:
                field_type_class = getattr(sys.modules[__name__], "FieldType")
                if issubclass(mod_class, field_type_class) is True:
                    class_name = getattr(mod_class, "name")
                    if class_name is None:
                        continue
                    field_map[class_name] = mod_class
        field_map = field_map

    return field_map


def get_class_by_type(class_name):
    get_field_type_map()
    if class_name in field_map:
        return field_map[class_name]
    raise ImportError("Field type class {} does not exists.".format(class_name))


def get_field_type_list():
    return list(get_field_type_map().keys())


def get_field_type_schema(field_type):
    get_field_type_map()

    def _expand_value_type(schema):

        allow_multiple = schema.get("allow_multiple", False)
        value_type = schema.get("value_type")
        # If the record doesn't set allow_multiple to True/False, allow the field to set the value.
        if issubclass(value_type, FieldType) is True:
            new_schema = value_type.schema
            return _expand_value_type(new_schema)
        elif issubclass(value_type, BaseEnum) is True:
            value = "<#ADD: " + value_type.build_example() + ">"
            return value
        elif issubclass(value_type, dict):
            value_block = {}
            for key, info in schema.get("schema").items():
                value_block[key] = _expand_value_type(info)
            return value_block
        else:
            value = "<#ADD: " + schema.get("desc", "Insert a {}".format(value_type.__name__)) + ">"
            if allow_multiple is True:
                value = [value]
            return value

    data = {
        "type": field_type
    }

    if field_type not in field_map:
        raise ValueError("Field type '{}' does not exists.".format(field_type))
    field_type_obj = field_map[field_type]
    field_schema = field_type_obj.schema
    data["value"] = _expand_value_type(field_schema)
    data["privacyScreen"] = field_schema.get("privacy_screen", False)
    field_type_obj.add_template_specifics(data)

    return data


class PasswordComplexity:

    def __init__(self, value=None, length=64, caps=0, lowercase=0, digits=0, special=0, filter_characters=None):

        # If a dictionary is passed in, use that get the attributes.
        if value is not None and type(value) is dict:
            length = value.get("length", length)
            caps = value.get("caps", caps)
            lowercase = value.get("lowercase", lowercase)
            digits = value.get("digits", digits)
            special = value.get("special", special)
            filter_characters = value.get("filter_characters", filter_characters)

        self.length = length
        self.caps = caps
        self.lowercase = lowercase
        self.digits = digits
        self.special = special

        self.filter_characters = filter_characters
        if isinstance(self.filter_characters, str):
            self.filter_characters = []
            self.filter_characters.extend(filter_characters)

    def to_dict(self):

        return {
            "length": self.length,
            "caps": self.caps,
            "lowercase": self.lowercase,
            "digits": self.digits,
            "special": self.special,
            "filter_characters": self.filter_characters
        }

    def replacement_char(self):

        """
        Get a replacement character that doesn't match the bad character.
        """

        new_char = None
        all_true = (self.lowercase + self.caps + self.digits + self.special) == 0

        attempt = 0
        while True:
            # If allow everything, then just get a lowercase letter
            if all_true is True:
                new_char = "abcdefghijklmnopqrstuvwxyz"[random.randint(0, 25)]

            # Else we need to find the first allowed character set.
            else:
                pick_one = random.randint(0, 3)
                if pick_one == 0 and self.lowercase > 0:
                    new_char = "abcdefghijklmnopqrstuvwxyz"[random.randint(0, 25)]
                if pick_one == 1 and self.caps > 0:
                    new_char = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"[random.randint(0, 25)]
                if pick_one == 2 and self.digits > 0:
                    new_char = "0123456789"[random.randint(0, 9)]
                if pick_one == 3 and self.special > 0:
                    new_char = "!@#$%^&*()"[random.randint(0, 9)]

                if new_char is None:
                    continue

            # If our new character is not in the list of bad characters, break out of the while
            if new_char not in self.filter_characters:
                break

            # Ok, some user might go crazy and filter out every letter, digit, and symbol and cause an invite loop.
            # If we can't find a good character after 25 attempts, error out.
            attempt += 1
            if attempt > 25:
                raise ValueError("Cannot filter character from password. The password complexity is too complex.")

        return new_char

    def filter_password(self, password):

        for bad_char in self.filter_characters:
            while bad_char in password:
                password = password.replace(bad_char, self.replacement_char(), 1)
        return password

    def generate_password(self):
        try:
            mod = import_module("keeper_secrets_manager_core.utils")
            password = getattr(mod, "generate_password")(
                length=self.length,
                lowercase=self.lowercase,
                uppercase=self.caps,
                digits=self.digits,
                special_characters=self.special
            )

            # If we have a character filter, remove bad characters
            if self.filter_characters is not None:
                password = self.filter_password(password)

        except ImportError as _:
            raise Exception("Cannot generate a random password. Requires keeper-secrets-manager-core module.")

        return password


class FieldType:
    schema = {"value_type": str}
    name = None

    # In a field that allows multiple FieldType, is there a "primary key" that makes a dictionary entry unique.
    # This use for Phones, which can many Phone entries, and if we need to figure out what is unique.
    group_key = None

    # Are multiple values allowed
    allow_multiple = False

    def __init__(self, *args, **kwargs):
        self._value = None

        # If the value is passed in, set it/overwrite the value in kwargs.
        if len(args) > 0:
            kwargs["value"] = args[0]

        # Get and remove the common keys
        value = kwargs.pop("value", None)
        label = kwargs.pop("label", None)
        required = kwargs.pop("required", None)
        privacy_screen = kwargs.pop("privacy_screen", None)

        # This will validate and set the value. If the value passed in is invalid, a ValueError exception is thrown.
        if value is not None:
            self.value = value

        self.label = label
        self.required = required
        self.privacy_screen = privacy_screen

        schema = self.get_schema()
        if schema.get("value_type") is dict:

            # Default to the passed in args for attribute variables. However if the value args exists, then use that
            # to get the attribute values.
            attr_dict = kwargs
            if self.value is not None:
                attr_dict = self.value[0]

            # This will create and set the attribute variables.
            for key in schema.get("schema", {}):
                setattr(self, key, attr_dict.get(key, None))

    def __str__(self):
        return f'{self.__class__.__name__}(label={self.label}, value={self.value}, required={self.required}, '\
               f'privacy_screen={self.privacy_screen})'

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        schema = self.get_schema()
        value_type = schema.get("value_type")

        if isinstance(val, list) is False:
            val = [val]

        # If the value of the field is another FieldType, convert raw value into an instance of that FieldType.
        if issubclass(value_type, FieldType) is True:
            for index in range(len(val)):
                if isinstance(val[index], value_type) is False:
                    # Create an instance, pass in value as the value and let the FieldType
                    val[index] = value_type(val[index])

        if self._is_valid(val) is True:
            self._value = val

    def add_value(self, new_value):
        val = self.value
        if val is None:
            val = []
        val.append(new_value)
        self.value = val

    def get_schema(self):
        schema = FieldType.schema
        if hasattr(self.__class__, "schema") is True:
            schema = self.__class__.schema

        return schema

    def get_type(self):
        return self.__class__.name

    @staticmethod
    def is_value_valid(value, schema, field_class):
        validator = schema.get("validate")
        value_type = schema.get("value_type")

        # The field value contains a dictionary of values.
        if value_type is dict:
            if type(value) is not dict:
                raise ValueError("The value is not a dictionary.")
            dict_schema = schema.get("schema")
            for key, info in dict_schema.items():
                # If the value is required, then make sure it exists and is not blank.
                if info.get("required", False) is True and (value.get(key) is None or value.get(key) == ""):
                    raise ValueError(f"The value key '{key}' is missing and it is required.")

                FieldType.is_value_valid(value.get(key), dict_schema.get(key), field_class)

        # The field value has an enumeration. Make sure the value is a valid value of the enumeration.
        elif value is not None and issubclass(value_type, BaseEnum):

            if value_type.enum_exists(value) is False:
                raise ValueError(f"The value of '{value}' for '{field_class}' is not valid.")

        # If the field value is a string, then it doesn't matter if an int, float, etc is passed in. However
        # we can do some validation on the value, so treat as string.
        elif value_type is str:
            if value is not None and validator is not None:
                value = str(value)
                validator_params = schema.get("validate_params", re.IGNORECASE)
                if re.match(str(validator), value, validator_params) is None:
                    raise ValueError(f"The value for' {field_class}' is not valid.")

        # Check if the value type is another FieldType, check the field type
        elif value is not None and isinstance(value_type, FieldType) is True:
            value_type.is_value_valid(value, schema, field_class)

    def _is_valid(self, value):
        schema = self.get_schema()
        for item in value:
            self.is_value_valid(item, schema, self.__class__.__name__)
        return True

    def build_value(self, schema, value=None, is_child:bool=False):

        value_type = schema.get("value_type")

        # If the value is an instance of FieldType
        if issubclass(value_type, FieldType) is True:
            return value.build_value(value_type.schema, None)

        # If the value type is an enumeration, get the value of the enumeration. The actual value may be
        # the enum or possible value. Find the right enum and get that enums value.
        elif issubclass(value_type, BaseEnum) is True:
            if value is None:
                return None
            return value_type.get_value(value)

        # If the value is a dictionary, then we are getting the values from the instance's attributes.
        elif value_type is dict:

            new_schema = schema.get("schema")
            value_dict = {}
            # Get the key and its schema.
            for key, info in new_schema.items():
                dict_value = None
                if not is_child:
                    # Get the value from the instance's attribute.
                    dict_value = self.build_value(info, getattr(self, key), is_child=True)
                else:
                    # Get value from dict /current value/
                    if value and key in value:
                        dict_value = self.build_value(info, value.get(key), is_child=True)

                if dict_value is not None:
                    value_dict[key] = dict_value

            return value_dict

        # At this point, we are a simple data type (99.999% we are str).
        else:
            if value is not None:
                if schema.get("format") is not None:
                    # This will take the value from the attribute and format, and validate it, into the desired
                    # format
                    value = self.format_value(
                        format_type=schema.get("format"),
                        value=value
                    )
            return value

    def to_dict(self):
        schema = self.get_schema()

        # Add the field camel case name of the class.
        field_dict = {
            "type": self.get_type()
        }
        # Add additional properties only if they are not blank. This is based on Vault UI behavior.
        if self.label is not None and self.label != "":
            field_dict["label"] = self.label
        if self.required is not None:
            field_dict["required"] = self.required
        if self.privacy_screen is not None:
            field_dict["privacyScreen"] = self.privacy_screen

        new_values = []

        # self.value is a list of values. We need to check them all.
        values = self.value
        if values is None:
            values = []

        # If the value is a dictionary, there is no value in self.value
        # since the value comes from the attributes. We need to fake values,
        # so set it one item of None. It won't be used, but a for loop will need it.
        if schema.get("value_type") is dict:
            values = [None]

        # Build add the values.
        for item in values:
            new_value = self.build_value(schema, item)
            self.is_value_valid(new_value, schema, self.__class__.__name__)
            new_values.append(new_value)

        field_dict["value"] = new_values

        return field_dict

    def to_json(self):
        return json.dumps(self.to_dict())

    @staticmethod
    def format_value(format_type, value):
        if hasattr(keeper_secrets_manager_helper.format, format_type) is True:
            return getattr(keeper_secrets_manager_helper.format, format_type)(value)
        raise ValueError("Could not find formatter {}".format(format_type))

    @staticmethod
    def add_template_specifics(data):
        pass

# -------------------------------------------------------------------------------------------------------------------


class Text(FieldType):
    name = "text"


class Url(FieldType):
    name = "url"


class PinCode(FieldType):
    name = "pinCode"


class Multiline(FieldType):
    name = "multiline"


# "file" - obsolete and removed legacy field - "fldt_file": { key: 'file_or_photo', default: "File or Photo" },
class FileRef(FieldType):
    name = "fileRef"
    # The validation checks to see if value is a Record UID.
    schema = {"value_type": str, "validate": UID_REGEX, "desc": "Record UID of File record."}


class Email(FieldType):
    name = "email"


class Phone(FieldType):
    name = "phoneItem"
    schema = {
        "value_type": dict,
        "schema": {
            "region": {"value_type": str, "desc": "Region"},
            "number": {"value_type": str, "desc": "Number"},
            "ext": {"value_type": str, "desc": "Extension"},
            "type": {"value_type": PhoneTypeEnum}
        }
    }


class Phones(FieldType):
    name = "phone"
    group_key = "type"
    allow_multiple = True
    schema = {"value_type": Phone}


class Name(FieldType):
    name = "name"
    schema = {
        "value_type": dict,
        "schema": {
            "first": {"value_type": str, "desc": "First Name"},
            "middle": {"value_type": str, "desc": "Middle name"},
            "last": {"value_type": str, "desc": "Last Name"}
        }
    }


class Address(FieldType):
    name = "address"
    schema = {
        "value_type": dict,
        "schema": {
            "street1": {"value_type": str, "desc": "Street"},
            "street2": {"value_type": str, "desc": "Street 2"},
            "city": {"value_type": str, "desc": "City"},
            "state": {"value_type": str, "desc": "State"},
            "zip": {"value_type": str, "desc": "Zip/Postal Code"},
            "country": {"value_type": CountryEnum, "desc": "ISO3166 Alpha-2 Country Code"},
        }
    }


class AddressRef(FieldType):
    name = "addressRef"
    # The validation checks to see if value is a Record UID.
    schema = {"value_type": str, "validate": UID_REGEX, "desc": "Record UID for Address record."}


class AccountNumber(FieldType):
    name = "accountNumber"


class Login(FieldType):
    name = "login"


class HiddenField(FieldType):
    name = "secret"


class Password(FieldType):
    name = "password"
    schema = {"value_type": str, "desc": "Password or Remove If Generating"}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._enforce_generation = None
        self._complexity = None

        self.enforce_generation = kwargs.pop("enforce_generation", None)
        self.complexity = kwargs.pop("complexity", None)
        self.password_generate = kwargs.pop("password_generate", None)

    def __str__(self):
        return f'{self.__class__.__name__}(label={self.label}, value={self.value}, required={self.required}, '\
               f'privacy_screen={self.privacy_screen}, enforce_generation={self.enforce_generation}'

    @property
    def enforce_generation(self):
        return self._enforce_generation

    @enforce_generation.setter
    def enforce_generation(self, value):
        if value is not None and type(value) is not bool:
            raise ValueError("enforce_generation needs to be a boolean value.")
        self._enforce_generation = value

    @property
    def complexity(self):
        return self._complexity

    @complexity.setter
    def complexity(self, value):
        if value is not None:
            if type(value) is dict:
                value = PasswordComplexity(value)
            if isinstance(value, PasswordComplexity) is not True:
                raise ValueError("complexity needs to be a PasswordComplexity instance.")
        self._complexity = value

    @staticmethod
    def add_template_specifics(data):
        data["enforceGeneration"] = False
        data["complexity"] = PasswordComplexity().to_dict()

    def to_dict(self):

        # If enforceGeneration is enabled in the record type or password_generate is True and the password has not
        # been set, then generate a password, ignore the value that is already there.
        if self.enforce_generation is True or (self.value is None and self.password_generate is True):
            self.generate_password()

        field_dict = super().to_dict()
        if self.enforce_generation is not None:
            field_dict["enforceGeneration"] = self.enforce_generation
        if self.complexity is not None:
            field_dict["complexity"] = self.complexity.to_dict()
        return field_dict

    def generate_password(self):
        if self.complexity is None:
            self.complexity = PasswordComplexity()
        self.value = [self.complexity.generate_password()]


class SecurityQuestions(FieldType):
    name = "securityQuestion"
    schema = {
        "value_type": dict,
        "schema": {
            "question": {"value_type": str, "desc": "Security Question"},
            "answer": {"value_type": str, "desc": "Answer To The Question"}
        }
    }


class OneTimePassword(FieldType):
    name = "otp"
    schema = {"value_type": str, "validate": r'^otpauth://', "desc": "URL starting with otpauth://"}


class OneTimeCode(FieldType):
    name = "oneTimeCode"
    schema = {"value_type": str, "validate": r'^otpauth://', "desc": "URL starting with otpauth://"}


class CardRef(FieldType):
    name = "cardRef"
    # The validation checks to see if value is a Record UID.
    schema = {"value_type": str, "validate": UID_REGEX, "desc": "Record UID of PaymentCard record."}


class PaymentCard(FieldType):
    name = "paymentCardItem"
    schema = {
        "value_type": dict,
        "schema": {
            "cardNumber": {"value_type": str, "desc": "Card Number"},
            "cardExpirationDate": {"value_type": str, "validate": r'^\d{2}\/\d{4}$',
                                   "desc": "Expiration Date as MM/YYYY"},
            "cardSecurityCode": {"value_type": str}
        }
    }


class PaymentCards(FieldType):
    name = "paymentCard"
    schema = {"value_type": PaymentCard}


class Date(FieldType):
    name = "date"
    schema = {"value_type": str, "format": "date_to_ms", "desc": "Date in ISO8601 Format or Epoch Milliseconds"}


class BirthDate(FieldType):
    name = "birthDate"
    schema = {"value_type": str, "format": "date_to_ms", "desc": "Birth Date in ISO8601 Format or Epoch Milliseconds"}


class ExpirationDate(FieldType):
    name = "expirationDate"
    schema = {"value_type": str, "format": "date_to_ms",
              "desc": "Expiration Date in ISO8601 Format or Epoch Milliseconds"}


class BankAccount(FieldType):
    name = "bankAccountItem"
    schema = {
        "value_type": dict,
        "schema": {
            "accountType": {"value_type": AccountTypeEnum},
            "otherType": {"value_type": str, "desc": "Other Type Description"},
            "routingNumber": {"value_type": str, "desc": "Routing Number"},
            "accountNumber": {"value_type": str, "desc": "Account Number"},
        }
    }


class BankAccounts(FieldType):
    name = "bankAccount"
    schema = {"value_type": BankAccount}


class KeyPair(FieldType):
    name = "keyPair"
    schema = {
        "value_type": dict,
        "schema": {
            "publicKey": {"value_type": str, "desc": "Public Key"},
            "privateKey": {"value_type": str, "desc": "Private Key. Normally a PEM file."},
        }
    }


class Host(FieldType):
    name = "host"
    schema = {
        "value_type": dict,
        "schema": {
            "hostName": {"value_type": str, "desc": "Hostname or IP"},
            "port": {"value_type": str, "desc": "Port"},
        }
    }


class LicenseNumber(FieldType):
    name = "licenseNumber"
    schema = {"value_type": str, "desc": "License Number"}


class SecureNote(FieldType):
    name = "note"
    schema = {"value_type": str, "desc": "Secret Note"}


class RecordRef(FieldType):
    name = "recordRef"
    # The validation checks to see if value is a Record UID.
    schema = {"value_type": str, "validate": UID_REGEX, "desc": "Record UID of the referenced record."}


class Schedule(FieldType):
    name = "schedule"
    schema = {
        "value_type": dict,
        "schema": {
            "type": {"value_type": str, "desc": "Type"},
            "cron": {"value_type": str, "desc": "Crontab format string"},
            # "utcTime" - replaced by time and tz
            "time": {"value_type": str, "desc": "Time"},
            "tz": {"value_type": str, "desc": "Time zone"},
            "weekday": {"value_type": str, "desc": "Day of the Week"},
            "intervalCount": {"value_type": int, "desc": "Interval Count"}
        }
    }


class DirectoryType(FieldType):
    name = "directoryType"
    schema = {"value_type": str, "desc": "Directory Type"}


class DatabaseType(FieldType):
    name = "databaseType"
    schema = {"value_type": str, "desc": "Database Type"}


class PamHostname(FieldType):
    name = "pamHostname"
    schema = {
        "value_type": dict,
        "schema": {
            "hostName": {"value_type": str, "desc": "Hostname or IP"},
            "port": {"value_type": str, "desc": "Port"},
        }
    }


class PamResources(FieldType):
    name = "pamResources"
    schema = {
        "value_type": dict,
        "schema": {
            "controllerUid": {"value_type": str, "desc": "Record UID of the Controller Record"},
            "folderUid": {"value_type": str, "desc": "Folder UID"},
            "resourceRef": {"value_type": list, "desc": "List with UIDs of resource records"},
            "allowedSettings": {
                "value_type": dict, "desc": "List with allowed settings flags",
                "schema": {
                    "connections": {"value_type": bool},
                    "portForwards": {"value_type": bool},
                    "rotation": {"value_type": bool},
                    "sessionRecording": {"value_type": bool},
                    "typescriptRecording": {"value_type": bool}
                }},
        }
    }


class Checkbox(FieldType):
    name = "checkbox"
    schema = {"value_type": bool, "desc": "Checkbox"}


class Passkey(FieldType):
    name = "passkey"
    schema = {
        "value_type": dict,
        "schema": {
            "privateKey": {"value_type": dict, "desc": "Private EC Key.",
            "schema": {
                "crv": {"value_type": str},
                "d": {"value_type": str},
                "ext": {"value_type": bool},
                "key_ops": {"value_type": list},
                "kty": {"value_type": str},
                "x": {"value_type": str},
                "y": {"value_type": str}
            }},
            "credentialId": {"value_type": str, "desc": "Credential Id"},
            "signCount": {"value_type": int, "desc": "Sign Count"},
            "userId": {"value_type": str, "desc": "User Id"},
            "relyingParty": {"value_type": str, "desc": "Relying Party"},
            "username": {"value_type": str, "desc": "User Name"},
            "createdDate": {"value_type": str, "format": "date_to_ms", "desc": "Creation Date in ISO8601 Format or Epoch Milliseconds"}
        }
    }


class Script(FieldType):
    name = "script"
    schema = {
        "value_type": dict,
        "schema": {
            "fileRef": {"value_type": str, "validate": UID_REGEX, "desc": "Record UID of File record."},
            "command": {"value_type": str, "desc": "Command"},
            "recordRef": {"value_type": str, "validate": UID_REGEX, "desc": "Record UID of the referenced record."}
        }
    }


class IsSSIDHidden(FieldType):
    name = "isSSIDHidden"
    schema = {"value_type": bool, "desc": "Is SSID Hidden"}


class WifiEncryption(FieldType):
    name = "wifiEncryption"


class Dropdown(FieldType):
    name = "dropdown"


class RbiUrl(FieldType):
    name = "rbiUrl"


class AppFiller(FieldType):
    name = "appFiller"
    schema = {
        "value_type": dict,
        "schema": {
            "applicationTitle": {"value_type": str, "desc": "Application Title"},
            "contentFilter": {"value_type": str, "desc": "Content Filter"},
            "macroSequence": {"value_type": str, "desc": "Macro Sequence"},
        }
    }


class PamRemoteBrowserSettings(FieldType):
    name = "pamRemoteBrowserSettings"
    schema = {
        "value_type": dict,
        "schema": {
            "connection": {
                "value_type": dict,
                "desc": "Connection details",
                "schema": {
                    "protocol": {"value_type": str},
                    "userRecords": {"value_type": list},
                    "allowUrlManipulation": {"value_type": bool},
                    "allowedUrlPatterns": {"value_type": str},
                    "allowedResourceUrlPatterns": {"value_type": str},
                    "httpCredentialsUid": {"value_type": str},
                    "autofillConfiguration": {"value_type": str}
                }
            },
        }
    }


class PamSettings(FieldType):
    name = "pamSettings"
    schema = {
        "value_type": dict,
        "schema": {
            "connection": {
                "value_type": dict,
                "desc": "PAM Settings",
                "schema": {
                    "portForward": {
                        "value_type": list,
                        "desc": "Port Settings",
                        "schema": {
                            "reusePort": {"value_type": bool},
                            "port": {"value_type": str},
                        }
                    },
                    "connection": {
                        "value_type": list,
                        "desc": "Connection Settings",
                        "schema": {
                            "protocol": {"value_type": str},
                            "userRecords": {"value_type": list},
                            "security": {"value_type": str},
                            "ignoreCert": {"value_type": bool},
                            "resizeMethod": {"value_type": str},
                            "colorScheme": {"value_type": str}
                        }
                    },
                }
            },
        }
    }


class TrafficEncryptionSeed(FieldType):
    name = "trafficEncryptionSeed"


# List of retired field types:
# trafficEncryptionKey - replaced by trafficEncryptionSeed
# pamProvider - deprecated for legacy/internal use only
# controller - deprecated for legacy/internal use only



# -*- coding: utf-8 -*-

from keeper_secrets_manager_helper.v3.field_type import FieldType, get_field_type_map
from keeper_secrets_manager_helper.field import Field, FieldSectionEnum
import re


class Parser:

    def __init__(self):
        # Find all the field type classes and create a map from the the camelcase name to the FieldType class
        self.field_map = get_field_type_map()

    def parse_field(self, field_args):

        field_objs = []

        # Might look like a list, but it could be a set, convert it
        if isinstance(field_args, set) is True:
            field_args = list(field_args)

        if isinstance(field_args, list) is False:
            field_args = [field_args]

        for arg in field_args:

            # First check if the field section is defined in the arg. The default is the standard fields, or 'f'
            field_section = FieldSectionEnum.STANDARD
            field_section_match = re.match(r'^(?P<field_section>.)\.', arg)
            if field_section_match is not None:
                field_section_value = field_section_match.group('field_section')
                if field_section_value == "f":
                    field_section = FieldSectionEnum.STANDARD
                elif field_section_value == "c":
                    field_section = FieldSectionEnum.CUSTOM
                else:
                    raise ValueError(f"Field section can only be 'f' or 'c'. The value '{field_section_value}' is not"
                                     "not valid.")
                # Remove the f. or c. from the arg
                arg = arg[2:]

            # At this point a camelcase field type name should be at the front of the arg, followed by a "[" or "="
            field_type_match = re.match(r'^(?P<field_type>\w+)[.\[=]', arg)
            if field_type_match is None:
                raise ValueError("Cannot find the field type name.")
            field_type = field_type_match.group("field_type")
            if field_type not in self.field_map:
                raise ValueError("Field type '{}' does not exists.".format(field_type))

            # This key designating what would make a complete field record unique. It used in grouping data into
            # a field record. It's like a primary key. See the Phones/Phone field type.
            group_key = self.field_map[field_type].group_key
            allow_multiple = self.field_map[field_type].allow_multiple

            # Remove the field type from the arg
            arg = arg[len(field_type):]

            field_label = None
            value_key = None

            while True:
                next_char = arg[0]
                arg = arg[1:]

                # Is the next thing a label for the field.
                if next_char == "[":

                    # Labels are weird. We need to character nibble this because the label contains a ]
                    # ie f.phone[\[BRACKETS\]]=... label = [BRACKETS]
                    # So we need to handle [, ], \ with special care.
                    index = 0
                    found_end = False
                    escape_mode = False
                    buffer = ""
                    while index < len(arg):
                        c = arg[index]
                        index += 1
                        # If we get the end ] and we are not in escape mode, we are done
                        if c == "]" and escape_mode is False:
                            found_end = True
                            break
                        # If we get a \ and we are not in escape mode, then turn on escape mode and disregard this char.
                        # This basically prevent use completing if we get a ]
                        elif c == "\\" and escape_mode is False:
                            escape_mode = True
                        else:
                            buffer += c
                            # Turn off escape mode since we got a character.
                            if escape_mode is True:
                                escape_mode = False

                    if found_end is False:
                        raise ValueError("Could not find the end of the label.")

                    if buffer != "":
                        field_label = buffer

                    # The + 1 is the end ]
                    arg = arg[index:]

                # If the next thing is a sub value of the value (ie street1 of Address)
                elif next_char == ".":

                    # If we have already gotten the sub value, then there is an error in the argument being parsed.
                    if value_key is not None:
                        raise ValueError(f"The value key '{value_key}' has already been found.")

                    # Get the text left of the =. Including the [ too.
                    sub_value_match = re.match(r'^(?P<sub_value_key>[\w_]+)[\[=.]', arg)
                    if sub_value_match is not None:
                        value_key = sub_value_match.group("sub_value_key")
                        schema = self.field_map[field_type].schema
                        value_type = schema.get("value_type")

                        # The value type might be another FieldType, get that FieldType's schema and value type
                        if issubclass(value_type, FieldType):
                            schema = value_type.schema
                            value_type = schema.get("value_type")

                        # If the value type is not a dictionary then we couldn't be using a value key. Throw
                        # an exception.
                        if value_type is not dict:
                            raise ValueError("The field type '{}' does not have value keys. "
                                             "Cannot set the value key {}.".format(field_type, value_key))

                        # If the value key is not in the dictionary, throw an exception.
                        schema = schema.get("schema")
                        if value_key not in schema:
                            raise ValueError("The field type '{}' does not have the value key '{}'. ".format(
                                field_type, value_key))

                        # The value key was found, use it.
                        arg = arg[len(value_key):]

                # Else we have the value
                else:
                    value = arg
                    break

            field_objs.append(
                Field(
                    field_section=field_section,
                    type=field_type,
                    label=field_label,
                    value_key=value_key,
                    group_key=group_key,
                    value=value,
                    allow_multiple=allow_multiple
                )
            )

        return field_objs



# -*- coding: utf-8 -*-

import enum


class BaseEnum(enum.Enum):

    @classmethod
    def enum_exists(cls, value):
        for e in cls:
            if e == value or (type(value) is str and e.value.lower() == value.lower()):
                return True
        return False

    @classmethod
    def get_value(cls, value):
        for e in cls:
            if e == value or (type(value) is str and e.value.lower() == value.lower()):
                return e.value
        raise ValueError(f"The value {value} is valid choice.")

    @classmethod
    def build_example(cls):
        values = []
        for e in cls:
            values.append(e.value)
        return " | ".join(sorted(values))


class PhoneTypeEnum(BaseEnum):
    MOBILE = "Mobile"
    HOME = "Home"
    WORK = "Work"


class AccountTypeEnum(BaseEnum):
    CHECKING = "Checking"
    SAVINGS = "Savings"
    OTHER = "Other"


# from iso3166 import countries
# for c in countries:
#    print("\t{} = \"{}\"".format(c.alpha2, c.alpha2))

# Was trying to dynamically create this enum, however too much work and it might not match
# the list on the Vault UI. So just list, so we can add missing one or remove ones that don't match
# the Vault


class CountryEnum(BaseEnum):
    AF = "AF"
    AX = "AX"
    AL = "AL"
    DZ = "DZ"
    AS = "AS"
    AD = "AD"
    AO = "AO"
    AI = "AI"
    AQ = "AQ"
    AG = "AG"
    AR = "AR"
    AM = "AM"
    AW = "AW"
    AU = "AU"
    AT = "AT"
    AZ = "AZ"
    BS = "BS"
    BH = "BH"
    BD = "BD"
    BB = "BB"
    BY = "BY"
    BE = "BE"
    BZ = "BZ"
    BJ = "BJ"
    BM = "BM"
    BT = "BT"
    BO = "BO"
    BQ = "BQ"
    BA = "BA"
    BW = "BW"
    BV = "BV"
    BR = "BR"
    IO = "IO"
    BN = "BN"
    BG = "BG"
    BF = "BF"
    BI = "BI"
    KH = "KH"
    CM = "CM"
    CA = "CA"
    CV = "CV"
    KY = "KY"
    CF = "CF"
    TD = "TD"
    CL = "CL"
    CN = "CN"
    CX = "CX"
    CC = "CC"
    CO = "CO"
    KM = "KM"
    CG = "CG"
    CD = "CD"
    CK = "CK"
    CR = "CR"
    CI = "CI"
    HR = "HR"
    CU = "CU"
    CW = "CW"
    CY = "CY"
    CZ = "CZ"
    DK = "DK"
    DJ = "DJ"
    DM = "DM"
    DO = "DO"
    EC = "EC"
    EG = "EG"
    SV = "SV"
    GQ = "GQ"
    ER = "ER"
    EE = "EE"
    ET = "ET"
    FK = "FK"
    FO = "FO"
    FJ = "FJ"
    FI = "FI"
    FR = "FR"
    GF = "GF"
    PF = "PF"
    TF = "TF"
    GA = "GA"
    GM = "GM"
    GE = "GE"
    DE = "DE"
    GH = "GH"
    GI = "GI"
    GR = "GR"
    GL = "GL"
    GD = "GD"
    GP = "GP"
    GU = "GU"
    GT = "GT"
    GG = "GG"
    GN = "GN"
    GW = "GW"
    GY = "GY"
    HT = "HT"
    HM = "HM"
    VA = "VA"
    HN = "HN"
    HK = "HK"
    HU = "HU"
    IS = "IS"
    IN = "IN"
    ID = "ID"
    IR = "IR"
    IQ = "IQ"
    IE = "IE"
    IM = "IM"
    IL = "IL"
    IT = "IT"
    JM = "JM"
    JP = "JP"
    JE = "JE"
    JO = "JO"
    KZ = "KZ"
    KE = "KE"
    KI = "KI"
    KP = "KP"
    KR = "KR"
    XK = "XK"
    KW = "KW"
    KG = "KG"
    LA = "LA"
    LV = "LV"
    LB = "LB"
    LS = "LS"
    LR = "LR"
    LY = "LY"
    LI = "LI"
    LT = "LT"
    LU = "LU"
    MO = "MO"
    MK = "MK"
    MG = "MG"
    MW = "MW"
    MY = "MY"
    MV = "MV"
    ML = "ML"
    MT = "MT"
    MH = "MH"
    MQ = "MQ"
    MR = "MR"
    MU = "MU"
    YT = "YT"
    MX = "MX"
    FM = "FM"
    MD = "MD"
    MC = "MC"
    MN = "MN"
    ME = "ME"
    MS = "MS"
    MA = "MA"
    MZ = "MZ"
    MM = "MM"
    NA = "NA"
    NR = "NR"
    NP = "NP"
    NL = "NL"
    NC = "NC"
    NZ = "NZ"
    NI = "NI"
    NE = "NE"
    NG = "NG"
    NU = "NU"
    NF = "NF"
    MP = "MP"
    NO = "NO"
    OM = "OM"
    PK = "PK"
    PW = "PW"
    PS = "PS"
    PA = "PA"
    PG = "PG"
    PY = "PY"
    PE = "PE"
    PH = "PH"
    PN = "PN"
    PL = "PL"
    PT = "PT"
    PR = "PR"
    QA = "QA"
    RE = "RE"
    RO = "RO"
    RU = "RU"
    RW = "RW"
    BL = "BL"
    SH = "SH"
    KN = "KN"
    LC = "LC"
    MF = "MF"
    PM = "PM"
    VC = "VC"
    WS = "WS"
    SM = "SM"
    ST = "ST"
    SA = "SA"
    SN = "SN"
    RS = "RS"
    SC = "SC"
    SL = "SL"
    SG = "SG"
    SX = "SX"
    SK = "SK"
    SI = "SI"
    SB = "SB"
    SO = "SO"
    ZA = "ZA"
    GS = "GS"
    SS = "SS"
    ES = "ES"
    LK = "LK"
    SD = "SD"
    SR = "SR"
    SJ = "SJ"
    SZ = "SZ"
    SE = "SE"
    CH = "CH"
    SY = "SY"
    TW = "TW"
    TJ = "TJ"
    TZ = "TZ"
    TH = "TH"
    TL = "TL"
    TG = "TG"
    TK = "TK"
    TO = "TO"
    TT = "TT"
    TN = "TN"
    TR = "TR"
    TM = "TM"
    TC = "TC"
    TV = "TV"
    UG = "UG"
    UA = "UA"
    AE = "AE"
    GB = "GB"
    US = "US"
    UM = "UM"
    UY = "UY"
    UZ = "UZ"
    VU = "VU"
    VE = "VE"
    VN = "VN"
    VG = "VG"
    VI = "VI"
    WF = "WF"
    EH = "EH"
    YE = "YE"
    ZM = "ZM"
    ZW = "ZW"



# -*- coding: utf-8 -*-

from importlib import import_module


class FieldType:

    default_version = "v3"

    @staticmethod
    def get_field_type_list(version=None):
        if version is None:
            version = FieldType.default_version
        mod = import_module(f"keeper_secrets_manager_helper.{version}.field_type")
        return getattr(mod, "get_field_type_list")()

    @staticmethod
    def get_field_type_schema(field_type, version=None):
        if version is None:
            version = FieldType.default_version
        mod = import_module(f"keeper_secrets_manager_helper.{version}.field_type")
        return getattr(mod, "get_field_type_schema")(field_type)



# -*- coding: utf-8 -*-

from .common import load_file
from importlib import import_module
import os
import re


class RecordType:

    default_version = "v3"

    @staticmethod
    def load_record_types(file, version=None):
        data = load_file(file)

        data_test = data
        if isinstance(data, list) is True:
            data_test = data[0]

        # Check if a structured file
        if data_test.get("kind") is not None:
            return RecordType.load_helper_record_types(data)
        else:
            if data_test.get("recordTypeId") is not None:
                return RecordType.load_commander_record_types(data, version)
            else:
                raise Exception("Cannot determine the type of record type file.")

    @staticmethod
    def load_helper_record_types(data):
        version = data.get("version")
        mod = import_module(f"keeper_secrets_manager_helper.{version}.record_type")
        return getattr(mod, "load_record_type_from_data")(data)

    @staticmethod
    def load_commander_record_types(data, version=None):
        if version is None:
            version = RecordType.default_version
        mod = import_module(f"keeper_secrets_manager_helper.{version}.record_type")
        return getattr(mod, "load_commander_record_type_from_data")(data)

    @staticmethod
    def find_and_load_record_type_schema_files(path):

        for root, dirs, files in os.walk(path):
            for file in files:
                if re.search("(json|ya*ml)$", file, re.IGNORECASE) is not None:
                    try:
                        RecordType.load_record_types(os.path.join(root, file))
                    except (Exception,):
                        pass

    @staticmethod
    def get_record_type_list(version=None):
        if version is None:
            version = RecordType.default_version
        mod = import_module(f"keeper_secrets_manager_helper.{version}.record_type")
        return getattr(mod, "get_record_type_list")()



# -*- coding: utf-8 -*-

from keeper_secrets_manager_helper.v3.field_type import FieldType, get_field_type_map
from keeper_secrets_manager_helper.v3.enum import BaseEnum
from keeper_secrets_manager_helper.common import load_file
import os
import yaml
import json
import re


# Make a base class for all our record types
class RecordType:

    def __init__(self):
        self.field_type_map = get_field_type_map()

    # self.schema and self.name works, however the dynamic class generation will add it. To make PyCharm happy,
    # get the schema and name using getattr.

    def get_schema(self):
        return getattr(self, "schema")

    def get_name(self):
        return getattr(self, "name")

    def get_standard_fields(self):
        return self.get_schema().get("fields", [])

    def get_custom_fields(self):
        return self.get_schema().get("custom_fields", [])

    def _expand_value_type(self, schema, allow_multiple=None, is_required=False):

        value_type = schema.get("value_type")
        # If the record doesn't set allow_multiple to True/False, allow the field to set the value.
        if allow_multiple is None:
            allow_multiple = schema.get("allow_multiple", False)
        if issubclass(value_type, FieldType) is True:
            new_schema = value_type.schema
            return self._expand_value_type(new_schema, is_required=is_required)
        elif issubclass(value_type, BaseEnum) is True:
            value = "<#ADD: " + value_type.build_example() + ">"
            return value
        elif issubclass(value_type, dict):
            value_block = {}
            for key, info in schema.get("schema").items():
                value_block[key] = self._expand_value_type(info, is_required=is_required)
            return value_block
        else:
            value = "<#ADD: " + schema.get("desc", "Insert a {}".format(value_type.__name__)) + ">"
            if allow_multiple is True:
                value = [value]
            return value

    def generate_template_dict(self, title=None, notes=None):
        fields = []

        for field in self.get_schema().get("fields"):

            field_type = field.get("type")

            data = {
                "type": field_type
            }

            if field_type not in self.field_type_map:
                raise ValueError("Field type '{}' does not exists.".format(field_type))
            field_type_obj = self.field_type_map[field_type]
            field_schema = field_type_obj.schema
            data["value"] = self._expand_value_type(field_schema,
                                                    allow_multiple=field.get("allow_multiple", None),
                                                    is_required=field.get("required", False))
            data["privacyScreen"] = field_schema.get("privacy_screen", False)
            field_type_obj.add_template_specifics(data)
            fields.append(data)

        template = {
            "version": "v3",
            "kind": "KeeperRecord",
            "data": [{
                "recordType": self.get_name(),
                "title": title if title is not None else "<#ADD: The title of record here. This is required.>",
                "notes": notes if notes is not None else "<#ADD: Add some notes or remove.>",
                "fields": fields,
            }]
        }

        return template

    @staticmethod
    def generate_yaml_template(template_dict):
        return yaml.dump(template_dict, sort_keys=False)

    @staticmethod
    def generate_json_template(template_dict):
        return json.dumps(template_dict, indent=4, sort_keys=False)

    def generate_template(self, output_format, title=None, notes=None):
        template_dict = self.generate_template_dict(title=title, notes=notes)
        if output_format == "json":
            return self.generate_json_template(template_dict)
        return self.generate_yaml_template(template_dict)


# This maps a record type/name to a type class
class_map_by_type = {}


def load_record_type_from_file(file):
    load_record_type_from_data(load_file(file))


def load_record_type_from_data(record_types):

    for item in record_types.get("data", []):

        # Make sure we haven't loaded this class already. Cannot overwrite classes.
        if item.get("name") in class_map_by_type:
            raise ValueError("Cannot overwrite class {}".format(item.get("class")))
        record_type_class = type(item.get("class"), (RecordType,), {
            "name": item.get("name"),
            "schema": {
                "fields": item.get("fields", [])
            }
        })
        globals()[record_type_class.__name__] = record_type_class
        class_map_by_type[item.get("name")] = record_type_class


default_record_type_file = "default_record_types.yml"

# Get the directory of the executable file. If last directory is keeper_secrets_manager_cli, get the parent
# directory. There is no keeper_secrets_manager_cli directory.

# This is the Pypi module installed check. The default_record_type_file.yml will be in this directory along
# the other V3 modules.
current_directory = os.path.dirname(__file__)
schema_dir = None
if os.path.exists(os.path.join(current_directory, default_record_type_file)) is True:
    schema_dir = current_directory

# Else this is the binary install. For PyInstaller, we move the YAML file to the rood
# directory of the application. We don't want to hardcode the directory, so walk backwards
# from here looking for the file. If we get to the root directory the file wasn't found.
else:
    # Change the filename to include the version
    default_record_type_file = f"v3_{default_record_type_file}"

    # Find the default_record_type_file in the path. Quit if reach the root directory.
    root_dir = os.path.abspath(os.sep)
    while current_directory != root_dir:
        current_directory = os.path.dirname(current_directory)
        if os.path.exists(os.path.join(current_directory, default_record_type_file)) is True:
            schema_dir = current_directory
            break
    if schema_dir is None:
        raise FileNotFoundError(f"Cannot find {default_record_type_file} in the binary app.")

load_record_type_from_file(os.path.join(schema_dir, default_record_type_file))


def get_class_by_type(class_name):
    if class_name in class_map_by_type:
        return class_map_by_type[class_name]
    raise ImportError("Record type class {} is not loaded.".format(class_name))


def get_record_type_list():
    return class_map_by_type.keys()


def make_class_name(name):
    name = name.lower()
    name = re.sub(r'[^a-zA-Z0-9\s]', '', name)
    name_parts = re.split(' +', name)
    name = ""
    for item in name_parts:
        if item is None or item == "":
            continue
        name += item[0].upper() + item[1:]
    return name


def camel_case_split(value):
    matches = re.finditer('.+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)', value)
    return "_".join([m.group(0) for m in matches]).lower()


def load_commander_record_type_from_file(file):
    load_commander_record_type_from_data(load_file(file))


def load_commander_record_type_from_data(data):

    if isinstance(data, dict) is True:
        data = [data]

    record_schema = {
        "data": []
    }

    index = 0
    for item in data:
        if item.get("recordTypeId") is None:
            raise ValueError(f"Missing recordTypeId for record index {index}. Is this an export of a Keeper "
                             f"Commander record type information.")
        if item.get("content") is None:
            raise ValueError(f"Missing content for record index {index}. Is this an export of a Keeper "
                             f"Commander record type information.")

        content = json.loads(item.get("content"))

        name = content.get('$id')
        class_name = make_class_name(name)

        record_data = {
            "class": class_name,
            "name": name,
            "fields": []
        }

        for field in content.get("fields"):
            field_data = {
                "type": field.get("$ref")
            }
            for param in ["label", "required", "privacyScreen", "enforceGeneration", "privacyScreen", "complexity"]:
                value = field.get(param)
                if value is not None:
                    field_data[camel_case_split(param)] = value

            record_data["fields"].append(field_data)

        record_schema["data"].append(record_data)

    load_record_type_from_data(record_schema)



# -*- coding: utf-8 -*-

from .exception import FileSyntaxException
import yaml
import json
import re
import os


def load_file(file):
    is_json = re.search("json$", file, re.IGNORECASE) is not None
    if os.path.exists(file) is False:
        raise Exception(f"Cannot find the file {file}")
    with open(file, 'r') as fh:
        if is_json is True:
            try:
                record_data = json.loads(fh.read())
            except json.JSONDecodeError as err:
                raise FileSyntaxException(f"The JSON had problems: {err.msg} around row "
                                          f"{err.lineno}, column {err.colno}")
        else:
            try:
                record_data = yaml.load(fh.read(), Loader=yaml.BaseLoader)
            except yaml.YAMLError as err:
                if hasattr(err, 'problem_mark'):
                    mark = err.problem_mark
                    raise FileSyntaxException(f"The YAML has problems around row "
                                              f"{mark.line + 1}, column {mark.column + 1}.")
                raise FileSyntaxException("The YAML has problems.")
    return record_data





# -*- coding: utf-8 -*-

import iso8601
from datetime import datetime


def date_to_ms(value):

    try:
        # Check if already an integer value
        try:
            value = int(value)
        # Else try to parse the string into a date
        except ValueError as _:
            dt = iso8601.parse_date(value)
            # Set the epoch timestamp to have the same timezone as the parsed value
            epoch = datetime.utcfromtimestamp(0).replace(tzinfo=dt.tzinfo)
            value = int((dt - epoch).total_seconds() * 1000.0)
    except iso8601.iso8601.ParseError as err:
        raise ValueError("Cannot format date/time as milliseconds: {}".format(err))

    return value



from setuptools import setup, find_packages
import os

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README.md file
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

install_requires = [
    'keeper-secrets-manager-core>=16.2.2',
    'pyyaml',
    'iso8601'
]

setup(
    name="keeper-secrets-manager-helper",
    version="1.0.5",
    description="Keeper Secrets Manager SDK helper for managing records.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Keeper Security",
    author_email="sm@keepersecurity.com",
    url="https://github.com/Keeper-Security/secrets-manager",
    license="MIT",
    keywords="Keeper Password Secrets Manager Helper Record",
    packages=find_packages(exclude=["tests", "tests.*"]),
    zip_safe=False,
    package_data={},
    include_package_data=True,
    install_requires=install_requires,
    python_requires='>=3.6',
    project_urls={
        "Bug Tracker": "https://github.com/Keeper-Security/secrets-manager/issues",
        "Documentation": "https://app.gitbook.com/"
                         "@keeper-security/s/secrets-manager/secrets-manager",
        "Source Code": "https://github.com/Keeper-Security/secrets-manager",
    },
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Security",
    ]
)
